from typing import Dict, Any
from django.conf import settings
from django.utils.safestring import mark_safe
from django.template.loader import get_template
from django.template import Node, Template, Variable, VariableDoesNotExist
from django.template.base import Token, Parser
import ast

from django_cotton.utils import ensure_quoted


class CottonIncompleteDynamicComponentError(Exception):
    """Raised when a dynamic component is missing required attributes."""


def cotton_component(parser: Parser, token: Token) -> "CottonComponentNode":
    """
    Template tag to render a cotton component with dynamic attributes.
    """
    bits = token.split_contents()
    if len(bits) < 3:
        raise ValueError("cotton_component tag requires at least a component path and key.")

    component_path = bits[1]
    component_key = bits[2]

    kwargs = {}
    for bit in bits[3:]:
        if "=" in bit:
            key, value = bit.split("=", 1)
        else:
            key, value = bit, ""
        kwargs[key] = value

    nodelist = parser.parse(("end_cotton_component",))
    parser.delete_first_token()

    return CottonComponentNode(nodelist, component_path, component_key, kwargs)


class CottonComponentNode(Node):
    def __init__(self, nodelist, component_path: str, component_key: str, kwargs: Dict[str, str]):
        self.nodelist = nodelist
        self.component_path = component_path
        self.component_key = component_key
        self.kwargs = kwargs

    def render(self, context) -> str:
        slot = self.nodelist.render(context)
        attrs = self._build_attrs(context)

        named_slots_ctx = self._process_named_slots(context)
        self._evaluate_expression_attributes(attrs, named_slots_ctx, context)

        template = self._get_template(context, attrs)
        attrs_string = self._build_attrs_string(attrs)
        accessible_attrs = {key.replace("-", "_"): value for key, value in attrs.items()}

        return self._render_template(
            template, context, slot, attrs, attrs_string, accessible_attrs, named_slots_ctx
        )

    def _build_attrs(self, context) -> Dict[str, Any]:
        attrs = {}
        for key, value in self.kwargs.items():
            value = self._strip_quotes(value)
            if key.startswith(":"):
                key = key[1:]
                attrs[key] = self._process_dynamic_attribute(key, value, context)
            elif value == "":
                attrs[key] = True
            else:
                attrs[key] = value
        return attrs

    @staticmethod
    def _strip_quotes(value: str) -> str:
        if value and value[0] == value[-1] and value[0] in ('"', "'"):
            return value[1:-1]
        return value

    def _process_dynamic_attribute(self, key: str, value: str, context) -> Any:
        try:
            return Variable(value).resolve(context)
        except VariableDoesNotExist:
            pass

        if value == "":
            return True

        value = self._parse_template_string(value, context)

        try:
            return ast.literal_eval(value)
        except (ValueError, SyntaxError):
            pass

        context.setdefault("ctn_unprocessable_dynamic_attrs", set()).add(key)
        return ""

    def _process_named_slots(self, context) -> Dict[str, Any]:
        all_named_slots_ctx = context.get("cotton_named_slots", {})
        named_slots_ctx = all_named_slots_ctx.get(self.component_key, {})
        all_named_slots_ctx[self.component_key] = {}
        return named_slots_ctx

    def _evaluate_expression_attributes(
        self, attrs: Dict[str, Any], named_slots_ctx: Dict[str, Any], context
    ) -> None:
        if "ctn_template_expression_attrs" in named_slots_ctx:
            for key in named_slots_ctx["ctn_template_expression_attrs"]:
                if key.startswith(":"):
                    evaluated = self._process_dynamic_attribute(key, named_slots_ctx[key], context)
                    key = key[1:]
                    attrs[key] = evaluated
                else:
                    attrs[key] = named_slots_ctx[key]

    def _get_template(self, context, attrs: Dict[str, Any]) -> Template:
        template_path = self._generate_component_template_path(attrs)
        cache = context.render_context.get(self)
        if cache is None:
            cache = context.render_context[self] = {}

        template = cache.get(template_path)
        if template is None:
            template = get_template(template_path)
            if hasattr(template, "template"):
                template = template.template
            cache[template_path] = template

        return template

    def _generate_component_template_path(self, attrs: Dict[str, Any]) -> str:
        if self.component_path == "component":
            if "is" not in attrs:
                raise CottonIncompleteDynamicComponentError(
                    'Cotton error: "<c-component>" should be accompanied by an "is" attribute.'
                )
            component_path = attrs["is"]
        else:
            component_path = self.component_path

        component_tpl_path = component_path.replace(".", "/").replace("-", "_")
        cotton_dir = getattr(settings, "COTTON_DIR", "cotton")
        return f"{cotton_dir}/{component_tpl_path}.html"

    @staticmethod
    def _build_attrs_string(attrs: Dict[str, Any]) -> str:
        return " ".join(f"{key}={ensure_quoted(value)}" for key, value in attrs.items())

    def _render_template(
        self,
        template: Template,
        context,
        slot: str,
        attrs: Dict[str, Any],
        attrs_string: str,
        accessible_attrs: Dict[str, Any],
        named_slots_ctx: Dict[str, Any],
    ) -> str:
        with context.push(
            {
                "slot": slot,
                "attrs_dict": attrs,
                "attrs": mark_safe(attrs_string),
                **accessible_attrs,
                **named_slots_ctx,
            }
        ):
            return template.render(context)

    @staticmethod
    def _parse_template_string(value: str, context) -> str:
        try:
            return Template(value).render(context)
        except (ValueError, SyntaxError):
            return value



from django import template
from django.utils.safestring import mark_safe


def cotton_slot(parser, token):
    """
    Template tag to render a cotton slot.
    is_expression_attr: bool whether we are using the named_slot as an "expression attribute" or not.
        expression attributes are from the use of native tags {{ {% and \n newlines within the attribute. We put them
        through as a named slot so they can be rendered and provided as a template variable
    """
    try:
        tag_name, slot_name, component_key, *optional = token.split_contents()
        is_expression_attr = optional[0] if optional else None
    except ValueError:
        raise template.TemplateSyntaxError("incomplete c-slot %r" % token.contents)

    nodelist = parser.parse(("end_cotton_slot",))
    parser.delete_first_token()
    return CottonSlotNode(slot_name, nodelist, component_key, is_expression_attr)


class CottonSlotNode(template.Node):
    def __init__(self, slot_name, nodelist, component_key, is_expression_attr):
        self.nodelist = nodelist
        self.slot_name = slot_name
        self.component_key = component_key
        self.is_expression_attr = is_expression_attr

    def render(self, context):
        output = self.nodelist.render(context)
        cotton_named_slots = context.setdefault("cotton_named_slots", {})
        component_slots = cotton_named_slots.setdefault(self.component_key, {})
        component_slots[self.slot_name] = mark_safe(output)

        # If the slot is being used to hold an expression attribute, we record it so it
        # can be transferred to attrs in the component
        if self.is_expression_attr:
            component_slots = context["cotton_named_slots"][self.component_key]
            expression_attrs = component_slots.setdefault("ctn_template_expression_attrs", [])
            expression_attrs.append(self.slot_name)

        return ""



from . import views
from django.urls import path
from django.views.generic import TemplateView

app_name = "django_cotton"


urlpatterns = [
    path("include", TemplateView.as_view(template_name="cotton_include.html")),
    path("test/compiled-cotton", views.compiled_cotton_test_view),
    path("test/native-extends", views.native_extends_test_view),
    path("test/native-include", views.native_include_test_view),
]



import warnings
import hashlib
import random
import os
import re

from django.template.loaders.base import Loader as BaseLoader
from django.core.exceptions import SuspiciousFileOperation
from django.template import TemplateDoesNotExist, Origin
from django.utils._os import safe_join
from django.template import Template
from django.apps import apps

from bs4 import BeautifulSoup, MarkupResemblesLocatorWarning
from bs4.formatter import HTMLFormatter

warnings.filterwarnings("ignore", category=MarkupResemblesLocatorWarning)


class Loader(BaseLoader):
    is_usable = True

    def __init__(self, engine, dirs=None):
        super().__init__(engine)
        self.cotton_compiler = CottonCompiler()
        self.cache_handler = CottonTemplateCacheHandler()
        self.dirs = dirs

    def get_contents(self, origin):
        cache_key = self.cache_handler.get_cache_key(origin)
        cached_content = self.cache_handler.get_cached_template(cache_key)

        if cached_content is not None:
            return cached_content

        template_string = self._get_template_string(origin.name)

        if "<c-" not in template_string and "{% cotton_verbatim" not in template_string:
            compiled = template_string
        else:
            compiled = self.cotton_compiler.process(template_string, origin.template_name)

        self.cache_handler.cache_template(cache_key, compiled)

        return compiled

    def get_template_from_string(self, template_string):
        """Create and return a Template object from a string. Used primarily for testing."""
        return Template(template_string, engine=self.engine)

    def _get_template_string(self, template_name):
        try:
            with open(template_name, "r", encoding=self.engine.file_charset) as f:
                return f.read()
        except FileNotFoundError:
            raise TemplateDoesNotExist(template_name)

    def get_dirs(self):
        """This works like the file loader with APP_DIRS = True."""
        dirs = self.dirs if self.dirs is not None else self.engine.dirs

        for app_config in apps.get_app_configs():
            template_dir = os.path.join(app_config.path, "templates")
            if os.path.isdir(template_dir):
                dirs.append(template_dir)

        return dirs

    def reset(self):
        """Empty the template cache."""
        self.cache_handler.reset()

    def get_template_sources(self, template_name):
        """Return an Origin object pointing to an absolute path in each directory
        in template_dirs. For security reasons, if a path doesn't lie inside
        one of the template_dirs it is excluded from the result set."""
        for template_dir in self.get_dirs():
            try:
                name = safe_join(template_dir, template_name)
            except SuspiciousFileOperation:
                # The joined path was located outside of this template_dir
                # (it might be inside another one, so this isn't fatal).
                continue

            yield Origin(
                name=name,
                template_name=template_name,
                loader=self,
            )


class UnsortedAttributes(HTMLFormatter):
    """This keeps BS4 from re-ordering attributes"""

    def attributes(self, tag):
        for k, v in tag.attrs.items():
            yield k, v


class CottonCompiler:
    DJANGO_SYNTAX_PLACEHOLDER_PREFIX = "__django_syntax__"
    COTTON_VERBATIM_PATTERN = re.compile(
        r"\{% cotton_verbatim %\}(.*?)\{% endcotton_verbatim %\}", re.DOTALL
    )
    DJANGO_TAG_PATTERN = re.compile(r"(\s?)(\{%.*?%\})(\s?)")
    DJANGO_VAR_PATTERN = re.compile(r"(\s?)(\{\{.*?\}\})(\s?)")
    HTML_ENTITY_PATTERN = re.compile(r"&[a-zA-Z]+;|&#[0-9]+;|&#x[a-fA-F0-9]+;")

    def __init__(self):
        self.django_syntax_placeholders = []
        self.html_entity_placeholders = []

    def process(self, content, template_name):
        content = self._replace_syntax_with_placeholders(content)
        content = self._replace_html_entities_with_placeholders(content)
        content = self._compile_cotton_to_django(content, template_name)
        content = self._fix_bs4_attribute_empty_attribute_behaviour(content)
        content = self._replace_placeholders_with_syntax(content)
        content = self._replace_placeholders_with_html_entities(content)
        content = self._remove_duplicate_attribute_markers(content)

        return content

    def _replace_html_entities_with_placeholders(self, content):
        """Replace HTML entities with placeholders so they dont get touched by BS4"""

        def replace_entity(match):
            entity = match.group(0)
            self.html_entity_placeholders.append(entity)
            return f"__HTML_ENTITY_{len(self.html_entity_placeholders) - 1}__"

        return self.HTML_ENTITY_PATTERN.sub(replace_entity, content)

    def _replace_placeholders_with_html_entities(self, content):
        for i, entity in enumerate(self.html_entity_placeholders):
            content = content.replace(f"__HTML_ENTITY_{i}__", entity)
        return content

    def _replace_syntax_with_placeholders(self, content):
        """Replace {% ... %} and {{ ... }} with placeholders so they dont get touched
        or encoded by bs4. We will replace them back after bs4 has done its job."""
        self.django_syntax_placeholders = []

        def replace_pattern(pattern, replacement_func):
            return pattern.sub(replacement_func, content)

        def replace_cotton_verbatim(match):
            """{% cotton_verbatim %} protects the content through the bs4 parsing process when we want to actually print
            cotton syntax in <pre> blocks."""
            inner_content = match.group(1)
            self.django_syntax_placeholders.append({"type": "verbatim", "content": inner_content})
            return (
                f"{self.DJANGO_SYNTAX_PLACEHOLDER_PREFIX}{len(self.django_syntax_placeholders)}__"
            )

        def replace_django_syntax(match):
            """Store if the match had at least one space on the left or right side of the syntax so we can restore it later"""
            left_space, syntax, right_space = match.groups()
            self.django_syntax_placeholders.append(
                {
                    "type": "django",
                    "content": syntax,
                    "left_space": bool(left_space),
                    "right_space": bool(right_space),
                }
            )
            return (
                f" {self.DJANGO_SYNTAX_PLACEHOLDER_PREFIX}{len(self.django_syntax_placeholders)}__ "
            )

        # Replace cotton_verbatim blocks
        content = replace_pattern(self.COTTON_VERBATIM_PATTERN, replace_cotton_verbatim)

        # Replace {% ... %}
        content = replace_pattern(self.DJANGO_TAG_PATTERN, replace_django_syntax)

        # Replace {{ ... }}
        content = replace_pattern(self.DJANGO_VAR_PATTERN, replace_django_syntax)

        return content

    def _compile_cotton_to_django(self, html_content, template_name):
        """Convert cotton <c-* syntax to {%."""
        soup = self._make_soup(html_content)

        # check if soup contains a 'c-vars' tag
        if cvars_el := soup.find("c-vars"):
            soup = self._wrap_with_cotton_vars_frame(soup, cvars_el)

        self._transform_components(soup, template_name)

        return str(soup.encode(formatter=UnsortedAttributes()).decode("utf-8"))

    def _replace_placeholders_with_syntax(self, content):
        """Replace placeholders with original syntax."""
        for i, placeholder in enumerate(self.django_syntax_placeholders, 1):
            if placeholder["type"] == "verbatim":
                placeholder_pattern = f"{self.DJANGO_SYNTAX_PLACEHOLDER_PREFIX}{i}__"
                content = content.replace(placeholder_pattern, placeholder["content"])
            else:
                """
                Construct the regex pattern based on original whitespace. This is to avoid unnecessary whitespace
                changes in the output that can lead to unintended tag type mutations,
                i.e. <div{% expr %}></div> --> <div__placeholder></div__placeholder> --> <div{% expr %}></div{% expr %}>
                """
                left_group = r"( ?)" if not placeholder["left_space"] else ""
                right_group = r"( ?)" if not placeholder["right_space"] else ""
                placeholder_pattern = (
                    f"{left_group}{self.DJANGO_SYNTAX_PLACEHOLDER_PREFIX}{i}__{right_group}"
                )

                content = re.sub(placeholder_pattern, placeholder["content"], content)

        return content

    def _remove_duplicate_attribute_markers(self, content):
        return re.sub(r"__COTTON_DUPE_ATTR__[0-9A-F]{5}", "", content, flags=re.IGNORECASE)

    def _fix_bs4_attribute_empty_attribute_behaviour(self, contents):
        """Bs4 adds ="" to valueless attribute-like parts in HTML tags that causes issues when we want to manipulate
        django expressions."""
        contents = contents.replace('=""', "")

        return contents

    def _wrap_with_cotton_vars_frame(self, soup, cvars_el):
        """If the user has defined a <c-vars> tag, wrap content with {% cotton_vars_frame %} to be able to create and
        govern vars and attributes. To be able to defined new vars within a component and also have them available in the
        same component's context, we wrap the entire contents in another component: cotton_vars_frame. Only when <c-vars>
        is present."""

        vars_with_defaults = []
        for var, value in cvars_el.attrs.items():
            # Attributes in context at this point will already have been formatted in _component to be accessible, so in order to cascade match the style.
            accessible_var = var.replace("-", "_")

            if value is None:
                vars_with_defaults.append(f"{var}={accessible_var}")
            elif var.startswith(":"):
                # If ':' is present, the user wants to parse a literal string as the default value,
                # i.e. "['a', 'b']", "{'a': 'b'}", "True", "False", "None" or "1".
                var = var[1:]  # Remove the ':' prefix
                accessible_var = accessible_var[1:]  # Remove the ':' prefix
                vars_with_defaults.append(f'{var}={accessible_var}|eval_default:"{value}"')
            else:
                # Assuming value is already a string that represents the default value
                vars_with_defaults.append(f'{var}={accessible_var}|default:"{value}"')

        cvars_el.decompose()

        # Construct the {% with %} opening tag
        opening = "{% cotton_vars_frame " + " ".join(vars_with_defaults) + " %}"
        opening = opening.replace("\n", "")
        closing = "{% endcotton_vars_frame %}"

        # Convert the remaining soup back to a string and wrap it within {% with %} block
        wrapped_content = (
            opening
            + str(soup.encode(formatter=UnsortedAttributes()).decode("utf-8")).strip()
            + closing
        )

        # Since we can't replace the soup object itself, we create new soup instead
        new_soup = self._make_soup(wrapped_content)

        return new_soup

    def _transform_components(self, soup, parent_key):
        """Replace <c-[component path]> tags with the {% cotton_component %} template tag"""
        for tag in soup.find_all(re.compile("^c-"), recursive=True):
            if tag.name == "c-slot":
                self._transform_named_slot(tag, parent_key)

                continue

            component_key = tag.name[2:]

            opening_tag = f"{{% cotton_component {component_key} {component_key} "

            # Store attributes that contain template expressions, they are when we use '{{' or '{%' in the value of an attribute
            expression_attrs = []

            # Build the attributes
            for key, value in tag.attrs.items():
                # BS4 stores class values as a list, so we need to join them back into a string
                if key == "class":
                    value = " ".join(value)

                # Django templates tags cannot have {{ or {% expressions in their attribute values
                # Neither can they have new lines, let's treat them both as "expression attrs"
                if self.DJANGO_SYNTAX_PLACEHOLDER_PREFIX in value or "\n" in value or "=" in value:
                    expression_attrs.append((key, value))
                    continue

                opening_tag += ' {}="{}"'.format(key, value)
            opening_tag += " %}"

            component_tag = opening_tag

            if expression_attrs:
                for key, value in expression_attrs:
                    component_tag += f"{{% cotton_slot {key} {component_key} expression_attr %}}{value}{{% end_cotton_slot %}}"

            if tag.contents:
                tag_soup = self._make_soup(tag.decode_contents(formatter=UnsortedAttributes()))
                self._transform_components(tag_soup, component_key)
                component_tag += str(
                    tag_soup.encode(formatter=UnsortedAttributes()).decode("utf-8")
                )

            component_tag += "{% end_cotton_component %}"

            # Replace the original tag with the compiled django syntax
            new_soup = self._make_soup(component_tag)
            tag.replace_with(new_soup)

        return soup

    def _transform_named_slot(self, slot_tag, component_key):
        """Compile <c-slot> to {% cotton_slot %}"""
        slot_name = slot_tag.get("name", "").strip()
        inner_html = "".join(str(content) for content in slot_tag.contents)

        # Check and process any components in the slot content
        slot_soup = self._make_soup(inner_html)
        self._transform_components(slot_soup, component_key)

        cotton_slot_tag = f"{{% cotton_slot {slot_name} {component_key} %}}{str(slot_soup.encode(formatter=UnsortedAttributes()).decode('utf-8'))}{{% end_cotton_slot %}}"
        slot_tag.replace_with(self._make_soup(cotton_slot_tag))

    def _make_soup(self, html):
        return BeautifulSoup(
            html,
            "html.parser",
            on_duplicate_attribute=self.handle_duplicate_attributes,
        )

    @staticmethod
    def handle_duplicate_attributes(tag_attrs, key, value):
        """BS4 cleans html and removes duplicate attributes. This would be fine if our target was html, but actually
        we're targeting Django Template Language. This contains expressions to govern content including attributes of
        any XML-like tag. It's perfectly fine to expect duplicate attributes per tag in DTL:

        <a href="#" {% if something %} class="this" {% else %} class="that" {% endif %}>Hello</a>

        The solution here is to make duplicate attribute keys unique across that tag so BS4 will not attempt to merge or
        replace existing. Then in post processing we'll remove the unique mask.
        """
        key_id = "".join(random.choice("0123456789ABCDEF") for i in range(5))
        key = f"{key}__COTTON_DUPE_ATTR__{key_id}"
        tag_attrs[key] = value


class CottonTemplateCacheHandler:
    """This mimics the simple template caching mechanism in Django's cached.Loader. Django's cached.Loader is a bit
    more performant than this one but it acts a decent fallback when the loader is not defined in the Django cache loader.

    TODO: implement integration to the cache backend instead of just memory. one which can also be controlled / warmed
    by the user and lasts beyond runtime restart.
    """

    def __init__(self):
        self.template_cache = {}

    def get_cached_template(self, cache_key):
        return self.template_cache.get(cache_key)

    def cache_template(self, cache_key, compiled_template):
        self.template_cache[cache_key] = compiled_template

    def get_cache_key(self, origin):
        try:
            cache_key = self.generate_hash([origin.name, str(os.path.getmtime(origin.name))])
        except FileNotFoundError:
            raise TemplateDoesNotExist(origin)

        return cache_key

    def generate_hash(self, values):
        return hashlib.sha1("|".join(values).encode()).hexdigest()

    def reset(self):
        self.template_cache.clear()



from django import template
from django.utils.safestring import mark_safe

from django_cotton.utils import ensure_quoted

register = template.Library()


def cotton_vars_frame(parser, token):
    """The job of the vars frame is:
    1. to filter out attributes declared as vars inside {{ attrs }} string.
    2. to provide default values to attributes.
    """
    bits = token.split_contents()[1:]  # Skip the tag name
    tag_kwargs = {}
    for bit in bits:
        key, value = bit.split("=")
        tag_kwargs[key] = parser.compile_filter(value)

    nodelist = parser.parse(("endcotton_vars_frame",))
    parser.delete_first_token()
    return CottonVarsFrameNode(nodelist, tag_kwargs)


class CottonVarsFrameNode(template.Node):
    def __init__(self, nodelist, kwargs):
        self.nodelist = nodelist
        self.kwargs = kwargs

    def render(self, context):
        provided_attrs = context.get("attrs_dict", {})
        unprocessable = context.get("ctn_unprocessable_dynamic_attrs", set())

        # Initialize vars based on the frame's kwargs and parent attrs
        c_vars = {}
        for key, value in self.kwargs.items():
            # Check if the var exists in component attrs;
            # We have an opinion here that if the provided attr value is empty, we should use the default value
            if key in provided_attrs and key not in unprocessable:
                c_vars[key] = provided_attrs[key]
            else:
                # Attempt to resolve each kwarg value (which may include template variables)
                c_vars[key] = value.resolve(context)

        # Excluding keys from {{ attrs }} that are identified as vars
        remaining_attrs_dict = {k: v for k, v in provided_attrs.items() if k not in c_vars}
        attrs_string = " ".join(f"{k}={ensure_quoted(v)}" for k, v in remaining_attrs_dict.items())
        accessible_names = {key.replace("-", "_"): value for key, value in c_vars.items()}

        tmp_context = {
            "attrs_dict": remaining_attrs_dict,
            "attrs": mark_safe(attrs_string),
            **accessible_names,
        }

        with context.push(**tmp_context):
            return self.nodelist.render(context)
